#!/usr/bin/env python3
#            _       _
#  _ __ ___ (_)_ __ (_)_ __ ___   ___  _ __
# | '_ ` _ \| | '_ \| | '_ ` _ \ / _ \| '_ \
# | | | | | | | | | | | | | | | | (_) | | | |
# |_| |_| |_|_|_| |_|_|_| |_| |_|\___/|_| |_|
#
# minimon - a minimal monitor
# Copyright (C) 2023 - Frans Fürst
#
# minimon is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# minimon is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details
# at <http://www.gnu.org/licenses/>.
#

""" Some little helper functions for checks
"""


def parse_df(lines):
    """
    Filesystem     1K-blocks      Used Available Use% Mounted on
    /dev/dm-0      981723644 814736524 117044544  88% /
    """
    return [
        f"{mountpoint} ({device}) {use}"
        for line in lines[1:]
        for elems in (line.split(),)
        if len(elems) > 5
        for mountpoint, device, use in ((elems[5], elems[0], int(elems[4][:-1])),)
        if use > 80
    ]


def parse_du(lines):
    """
    USER         PID %CPU %MEM    VSZ   RSS TTY      STAT START   TIME COMMAND
    root           1  0.0  0.0 173724  9996 ?        Ss   Aug13   0:04 /usr/lib/systemd/systemd rhgb --switched-root --system --deserialize 31
    """
    return [
        f"{user} {cpu} {mem} {cmd.split(maxsplit=1)[0]}"
        for line in lines[1:]
        for elems in (line.split(maxsplit=10),)
        if len(elems) > 10
        for user, cpu, mem, cmd in ((elems[0], float(elems[2]), float(elems[3]), elems[10]),)
        if cpu > 10
    ]



#!/usr/bin/env python3
#            _       _
#  _ __ ___ (_)_ __ (_)_ __ ___   ___  _ __
# | '_ ` _ \| | '_ \| | '_ ` _ \ / _ \| '_ \
# | | | | | | | | | | | | | | | | (_) | | | |
# |_| |_| |_|_|_| |_|_|_| |_| |_|\___/|_| |_|
#
# minimon - a minimal monitor
# Copyright (C) 2023 - Frans Fürst
#
# minimon is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# minimon is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details
# at <http://www.gnu.org/licenses/>.
#

"""The core stuff"""

import asyncio
import contextvars
import logging
import signal
import threading
from collections.abc import Callable, MutableMapping

from rich.logging import RichHandler
from textual import on
from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.message import Message
from textual.widgets import Footer, Header, Log, Pretty, RichLog, Static

from minimon.logging_utils import setup_logging


def log() -> logging.Logger:
    """Logger for this module"""
    return logging.getLogger("minimon")


class Context:
    def __init__(self):
        self.things = {}
        self.log_widgets = {}
        self.logger_context = contextvars.ContextVar("logger_context")

    def add(self, name, generator):
        self.things[name] = generator

    def current_logger(self):
        return self.log_widgets[self.logger_context.get()]

    def set_current_logger(self, name, log_widget):
        self.logger_context.set(name)
        self.log_widgets[name] = log_widget


class RichLogHandler(RichHandler):
    def __init__(self, widget: RichLog):
        super().__init__(
            # show_time = False,
            # omit_repeated_times = False,
            # show_level = False,
            show_path=False,
            # enable_link_path = False,
            markup=False,
            # rich_tracebacks = False,
            # tracebacks_width: Optional[int] = None,
            # tracebacks_extra_lines: int = 3,
            # tracebacks_theme: Optional[str] = None,
            # tracebacks_word_wrap: bool = True,
            # tracebacks_show_locals: bool = False,
            # tracebacks_suppress: Iterable[Union[str, ModuleType]] = (),
            # locals_max_length: int = 10,
            # locals_max_string: int = 80,
            # log_time_format = "[%x %X]",
            # keywords= None,
        )
        self.widget: RichLog = widget

    def emit(self, record: logging.LogRecord) -> None:
        message = self.format(record)
        message_renderable = self.render_message(record, message)
        traceback = None
        log_renderable = self.render(
            record=record, traceback=traceback, message_renderable=message_renderable
        )
        self.widget.write(log_renderable)


class TaskWidget(Static):
    def compose(self) -> ComposeResult:
        yield Pretty({}, classes="box")
        yield Log(classes="box")


class MiniMoni(App[None]):
    """Terminal monitor for minimon"""

    CSS_PATH = "minimon.css"

    def __init__(self, context) -> None:
        super().__init__()
        self._widget_container = VerticalScroll()
        self.context = context

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Footer()
        yield self._widget_container
        yield RichLog(
            # max_lines=None,
            # min_width=78,
            # wrap=False,
            # highlight=True,
            # markup=True,
            # auto_scroll=True,
            # name=None,
            # id=None,
            # classes=None,
            # disabled=False
        )

    @staticmethod
    async def task(name, task_fn, context: Context, widget: TaskWidget, update, cleanup):
        try:
            log().info("task %r started", task_fn)
            context.set_current_logger(name, widget.query_one(Log))
            async for data in task_fn:
                update(data)

        except Exception:
            log().exception("exception in %r", task_fn)
        finally:
            log().info("task %r terminated", task_fn)
            await cleanup()

    async def create_widget(self, title) -> TaskWidget:
        await self._widget_container.mount(widget := TaskWidget())
        widget.border_title = title
        return widget

    async def remove_widget(self, widget: TaskWidget) -> None:
        await widget.remove()

    async def add_task(self, name, task) -> TaskWidget:
        widget = await self.create_widget(name)
        asyncio.ensure_future(
            self.task(
                name,
                task,
                self.context,
                widget,
                widget.query_one(Pretty).update,
                lambda: self.remove_widget(widget),
            )
        )
        return widget

    async def on_mount(self) -> None:
        """UI entry point"""

        log().handlers = [RichLogHandler(self.query_one(RichLog))]

        for name, generator in self.context.things.items():
            await self.add_task(name, generator)

    @on(Message)
    async def on_msg(self, *event: str) -> None:
        """Generic message handler"""
        # log().debug("Event: %s", event)

    async def exe(self, on_exit: Callable[[], None]) -> None:
        """Execute and quit application"""
        try:
            await self.run_async()
        finally:
            on_exit()


def terminate(terminator: threading.Event) -> None:
    """Sends a signal to async tasks to tell them to stop"""
    try:
        terminator.set()
        for task in asyncio.all_tasks():
            task.cancel()
        asyncio.get_event_loop().stop()
    except Exception:  # pylint: disable=broad-except
        log().exception("terminate:")


def install_signal_handler(loop: asyncio.AbstractEventLoop, on_signal: Callable[[], None]) -> None:
    """Installs the CTRL+C application termination signal handler"""
    for signal_enum in [signal.SIGINT, signal.SIGTERM]:
        loop.add_signal_handler(signal_enum, on_signal)


class Singleton(type):
    """Yes, a Singleton"""

    _instances: MutableMapping[type, object] = {}

    def __call__(cls: "Singleton", *args: object, **kwargs: object) -> object:
        """Creates an instance if not available yet, returns it"""
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


def serve(context: Context, log_level="INFO") -> None:
    """Synchronous entry point"""
    setup_logging(log_level)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    terminator = threading.Event()

    asyncio.ensure_future(MiniMoni(context).exe(lambda: terminate(terminator)))

    try:
        install_signal_handler(loop, lambda: terminate(terminator))
        log().info("CTRL+C to quit")
        loop.run_forever()
    finally:
        log().debug("finally - loop.run_forever()")



#!/usr/bin/env python3

""" Monitor my private infrastructure
"""

from minimon.builder import Host, Monitor, iterate, process_output, view
from minimon.plugins import parse_df, parse_du

hosts = (
    Host("localhost", ssh_name="root"),
    Host("om-office.de", ssh_name="frans", ssh_port=2222),
    # Host("zentrale", ssh_name="frans"),
    # Host("remarkable", ssh_name="frans"),
    # Host("handy", ssh_name="frans"),
)

with Monitor("Private inf"):

    @view("host", hosts)
    async def local_resources(host: Host):
        """Provides quick summary of system sanity"""
        state = {}
        async for name, lines in iterate(
            ps=process_output(host, "ps wauxw", "1"),
            df=process_output(host, "df -P", "2"),
        ):
            state[name] = parse_df(lines) if name == "df" else parse_du(lines)
            yield state



#!/usr/bin/env python3

"""minimon command line interface
"""

from argparse import ArgumentParser
from argparse import Namespace as Args
from pathlib import Path

import chime  # type: ignore[import]

from minimon.server import Context, serve


def parse_args() -> Args:
    """Cool git like multi command argument parser"""
    parser = ArgumentParser()
    parser.add_argument(
        "--log-level",
        "-l",
        choices=["ALL_DEBUG", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        help="Sets the logging level - ALL_DEBUG sets all other loggers to DEBUG, too",
        type=str.upper,
        default="INFO",
    )
    parser.set_defaults(func=lambda *_: parser.print_usage())
    subparsers = parser.add_subparsers(help="available commands", metavar="CMD")

    parser_serve = subparsers.add_parser("serve")
    parser_serve.set_defaults(func=fn_serve)

    return parser.parse_args()


def fn_serve(args: Args) -> None:
    """Entry point for event consistency check"""
    chime.theme("big-sur")
    chime.notify_exceptions()
    serve(Context(), log_level=args.log_level).serve()


def main() -> int:
    """Entry point for everything else"""
    (args := parse_args()).func(args)
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

""" Monitor Checkmk CI

local
 - diskspace / cpu

remote: fra / win / nexus / ci / tstbuild / review
 - diskspace / cpu

fra
 - docker
   - total cpu / ram
   - number of containers
   - per container
       - start / stop
       - cpu / ram
       - volumes
       - associated jenkins job
       -
   - number of images
   - number of volumes

jenkins
   - job tree
   - warning about certain job results

nexus

actions
 - rebuild
 - kill/delete containers/volumes/tags/images
 - open/close branches

"""

from minimon.builder import Host, Monitor, iterate, process_output, view
from minimon.plugins import parse_df, parse_du

hosts = (
    Host("localhost", ssh_name="root"),
    Host("build-fra-001", ssh_name="root"),
    Host("build-fra-002", ssh_name="root"),
    Host("build-fra-003", ssh_name="root"),
    Host("ci", ssh_name="root"),
    Host("review", ssh_name="root"),
    Host("artifacts", ssh_name="root"),
    Host("bazel-cache", ssh_name="root"),
    Host("tstbuilds-artifacts", ssh_name="root"),
    Host("devpi", ssh_name="root"),
)

with Monitor("ci_dashboard"):

    @view("host", hosts)
    async def local_resources(host):
        state = {}
        async for name, lines in iterate(
            ps=process_output(host, "ps wauxw", "5"),
            df=process_output(host, "df -P", "60"),
        ):
            state[name] = parse_df(lines) if name == "df" else parse_du(lines)
            yield state



#!/usr/bin/env python3
#            _       _
#  _ __ ___ (_)_ __ (_)_ __ ___   ___  _ __
# | '_ ` _ \| | '_ \| | '_ ` _ \ / _ \| '_ \
# | | | | | | | | | | | | | | | | (_) | | | |
# |_| |_| |_|_|_| |_|_|_| |_| |_|\___/|_| |_|
#
# minimon - a minimal monitor
# Copyright (C) 2023 - Frans Fürst
#
# minimon is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# minimon is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details
# at <http://www.gnu.org/licenses/>.
#

import asyncio
import functools
import logging
import sys
from asyncio.subprocess import PIPE, create_subprocess_exec
from collections.abc import Callable, Sequence
from contextlib import suppress
from dataclasses import dataclass
from functools import partial
from itertools import count

import asyncssh

from minimon.server import Context, Singleton, serve


def log() -> logging.Logger:
    """Logger for this module"""
    return logging.getLogger("minimon")


class GlobalMonitorContext(Context, metaclass=Singleton):
    ...


@dataclass
class Host:
    name: str
    ip_address: None | str = None
    ssh_name: None | str = None
    ssh_key_file: None | str = None
    ssh_key_passphrase_cmd: None | str = None
    ssh_port: None | int = None

    def __str__(self):
        return self.name


Hosts = Sequence[Host]

# https://realpython.com/primer-on-python-decorators/#syntactic-sugar
def view(arg_name, arg_values):
    def decorator_view(afunc: Callable[[], int]):
        @functools.wraps(afunc)
        async def wrapper_view(*args: object, **kwargs: object) -> object:
            generator = afunc(*args, **kwargs)
            while True:
                try:
                    yield await anext(generator)
                except StopAsyncIteration:
                    log().info(f"StopAsyncIteration in {afunc.__name__}")
                    break
                except Exception:
                    log().exception("Unhandled exception in view generator:")

        fn_name = afunc.__name__
        for arg_value in arg_values:
            GlobalMonitorContext().add(
                f"{fn_name}-{arg_value}", wrapper_view(**{arg_name: arg_value})
            )

    return decorator_view


class HostConnection:
    def __init__(self, host: Host, log_fn) -> None:
        self.host_info = host
        self.log_fn = log_fn
        self.ssh_connection: None | asyncssh.SSHClientConnection = None

    async def __aenter__(self) -> "HostConnection":
        if self.host_info.ssh_name:
            self.ssh_connection = await asyncssh.connect(
                self.host_info.ip_address or self.host_info.name,
                port=self.host_info.ssh_port or (),
                username=self.host_info.ssh_name,
                # client_keys=[pkey],
            )

        return self

    async def __aexit__(self, *args: object) -> bool:
        if self.ssh_connection:
            self.ssh_connection.close()
        return True

    async def execute(self, command):
        if self.ssh_connection:

            def clean_lines(raw_line: str, log_widget) -> str:
                line = raw_line.strip("\n")
                log_widget.write_line(line)
                return line

            process = await self.ssh_connection.create_process(command)
            stdout, stderr, return_code = await asyncio.gather(
                self.log_fn(process.stdout, clean_lines),
                self.log_fn(process.stderr, clean_lines),
                process.wait(),
            )
            return stdout, stderr, return_code.returncode
        else:

            def clean_bytes(raw_line: bytes, log_widget) -> str:
                line = raw_line.decode().strip("\n")
                log_widget.write_line(line)
                return line

            process = await create_subprocess_exec(command, stdout=PIPE, stderr=PIPE)
            stdout, stderr, return_code = await asyncio.gather(
                self.log_fn(process.stdout, clean_bytes),
                self.log_fn(process.stderr, clean_bytes),
                process.wait(),
            )
            return stdout, stderr, return_code


async def process_output(host: Host, command, when: str):
    async def listen(log_widget, stream, clean_fn):
        return [clean_fn(raw_line, log_widget) async for raw_line in stream]

    logger = GlobalMonitorContext().current_logger()
    iterations = None

    async with HostConnection(host, partial(listen, logger)) as connection:
        for iteration in count():
            if iterations is not None and iteration >= iterations:
                break

            log().info("start task %r: %d", command, iteration)

            stdout, stderr, return_code = await connection.execute(command)

            log().info("task %r: %d, returned %d", command, iteration, return_code)

            yield stdout

            await asyncio.sleep(int(when))


async def iterate(**generator_defs):
    async def bundle(coro):
        async for result in coro:
            yield coro, result

    def task_from(name, coro):
        return asyncio.create_task(anext(bundle(coro)), name=name)

    tasks = set(task_from(name, coro) for name, coro in generator_defs.items())

    while tasks:
        done, tasks = await asyncio.wait(fs=tasks, return_when=asyncio.FIRST_COMPLETED)
        for event in done:
            with suppress(StopAsyncIteration):
                coro, result = event.result()
                name = event.get_name()
                tasks.add(task_from(name, coro))
                yield name, result


class Monitor:
    def __init__(self, name: str, log_level="INFO") -> None:
        self.name = name
        self.log_level = log_level

    def __enter__(self) -> "Monitor":
        return self

    def __exit__(self, *args: object) -> bool:
        if sys.exc_info() != (None, None, None):
            raise

        serve(GlobalMonitorContext(), self.log_level)
        return True


# with suppress(FileNotFoundError):
# with open(CONFIG_FILE) as config_file:
# config = yaml.load(config_file, yaml.Loader)
# print(config)



#!/usr/bin/env python3

"""Common stuff shared among modules"""

import logging
import sys
import threading
import traceback
from collections.abc import Iterator, Mapping
from types import TracebackType

LOG_LEVELS = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")


def log() -> logging.Logger:
    """Logger for this module"""
    return logging.getLogger("minimon")


def stack_str(depth: int = 0) -> str:
    """Create a tiny string from current call stack"""

    def stack_fns() -> Iterator[str]:
        # pylint: disable=protected-access
        stack = list(reversed(traceback.extract_stack(sys._getframe(depth))))
        for site in stack:
            if site.filename != stack[0].filename or site.name == "<module>":
                break
            yield site.name

    return ">".join(reversed(list(stack_fns())))


def setup_logging(level: str = "INFO") -> None:
    """Make logging fun"""

    class CustomLogger(logging.Logger):
        """Logs a record the way we want it"""

        # pylint: disable=too-many-arguments
        def makeRecord(
            self,
            name: str,
            level: int,
            fn: str,
            lno: int,
            msg: object,
            args: tuple[object, ...] | Mapping[str, object],
            exc_info: tuple[type[BaseException], BaseException, TracebackType | None]
            | tuple[None, None, None]
            | None,
            func: str | None = None,
            extra: Mapping[str, object] | None = None,
            sinfo: str | None = None,
        ) -> logging.LogRecord:
            """Creates a log record with a 'stack' element"""
            new_extra = {
                **(extra or {}),
                **{
                    "stack": stack_str(5),
                    "posixTID": threading.get_native_id(),
                },
            }
            return super().makeRecord(
                name, level, fn, lno, msg, args, exc_info, func, new_extra, sinfo
            )

    # for lev in LOG_LEVELS:
    #    logging.addLevelName(getattr(logging, lev), f"{lev[0] * 2}")

    logging.setLoggerClass(CustomLogger)

    log().setLevel(getattr(logging, level.split("_")[-1]))
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(getattr(logging, level.split("_")[-1]))
    stream_handler.setFormatter(
        logging.Formatter(
            "(%(levelname)s) %(asctime)s | %(posixTID)d | %(stack)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    log().handlers = [stream_handler]
    logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)

    # https://stackoverflow.com/questions/76788727/how-can-i-change-the-debug-level-and-format-for-the-quart-i-e-hypercorn-logge
    # https://pgjones.gitlab.io/hypercorn/how_to_guides/logging.html#how-to-log
    # https://www.phind.com/agent?cache=clkqhh48y001smg0832tvq1rl

    # from quart.logging import default_handler
    # logging.getLogger('quart.app').removeHandler(default_handler)
    # logger = logging.getLogger("hypercorn.error")
    # logger.removeHandler(default_handler)
    # logger.addHandler(ch)
    # logger.setLevel(logging.WARNING)
    # logger.propagate = False
